local ____lualib = require("lualib_bundle")
local __TS__Class = ____lualib.__TS__Class
local Map = ____lualib.Map
local __TS__ClassExtends = ____lualib.__TS__ClassExtends
local __TS__TypeOf = ____lualib.__TS__TypeOf
local __TS__New = ____lualib.__TS__New
local ____exports = {}
local ____types = require("functions.types")
local isFunction = ____types.isFunction
local isPrimitive = ____types.isPrimitive
--- `DefaultMap` is a data structure that makes working with default values easier. It extends a
-- `Map` and adds additional methods.
-- 
-- It is a common pattern to look up a value in a `Map`, and then, if the value does not exist, set
-- a default value for the key, and then return the default value. `DefaultMap` abstracts this
-- operation away by providing the `getAndSetDefault` method.
-- 
-- Using a `DefaultMap` is nice because it makes code more declarative, since you specify what the
-- default value is alongside the types of the keys/values.
-- 
-- When instantiating a new `DefaultMap`, you must specify default value as the first argument. (The
-- default value is the initial value that will be assigned to every new entry in the
-- `getAndSetDefault` method.) For example:
-- 
-- ```ts
-- // Initializes a new empty DefaultMap with a default value of "foo".
-- const defaultMapWithString = new DefaultMap<string, string>("foo");
-- 
-- const value = defaultMapWithString.getAndSetDefault("bar");
-- // value is now "foo" and an entry for "bar" is now set.
-- ```
-- 
-- Sometimes, instead of having a static initial value for every entry in the map, you will want a
-- dynamic initial value that is contingent upon the key or some other variable. In these cases, you
-- can instead specify that the `DefaultMap` should run a function that will return the initial
-- value. (This is referred to as a "factory function".) For example:
-- 
-- ```ts
-- // Initializes a new empty DefaultMap with a default value based on "someGlobalVariable".
-- const factoryFunction = () => someGlobalVariable ? 0 : 1;
-- const defaultMapWithFactoryFunction = new DefaultMap<string, string>(factoryFunction);
-- ```
-- 
-- Note that in TypeScript and Lua, booleans, numbers, and strings are "passed by value". This means
-- that when the `DefaultMap` creates a new entry, if the default value is one of these 3 types, the
-- values will be copied. On the other hand, arrays and maps and other complex data structures are
-- "passed by reference". This means that when the `DefaultMap` creates a new entry, if the default
-- value is an array, then it would not be copied. Instead, the same shared array would be assigned
-- to every entry. Thus, to solve this problem, any variable that is passed by reference must be
-- created using a factory function to ensure that each copy is unique. For example:
-- 
-- ```ts
-- // Initializes a new empty DefaultMap with a default value of a new empty array.
-- const factoryFunction = () => [];
-- const defaultMapWithArray = new DefaultMap<string, string[]>(factoryFunction);
-- ```
-- 
-- In the previous two examples, the factory functions did not have any arguments. But you can also
-- specify a factory function that takes one or more arguments:
-- 
-- ```ts
-- const factoryFunction = (arg: boolean) => arg ? 0 : 1;
-- const defaultMapWithArg = new DefaultMap<string, string, [arg: boolean]>(factoryFunction);
-- ```
-- 
-- Similar to a normal `Map`, you can also include an initializer list in the constructor as the
-- second argument:
-- 
-- ```ts
-- // Initializes a DefaultMap with a default value of "foo" and some initial values.
-- const defaultMapWithInitialValues = new DefaultMap<string, string>("foo", [
--   ["a1", "a2"],
--   ["b1", "b2"],
-- ], );
-- ```
-- 
-- Finally, note that `DefaultMap` has the following additional utility methods:
-- 
-- - `getAndSetDefault` - The method that is called inside the overridden `get` method. In most
--   cases, you can use the overridden `get` method instead of calling this function directly.
--   However, if a factory function was provided during instantiation, and the factory function has
--   one or more arguments, then you must call this method instead (and provide the corresponding
--   arguments).
-- - `getDefaultValue` - Returns the default value to be used for a new key. (If a factory function
--   was provided during instantiation, this will execute the factory function.)
-- - `getConstructorArg` - Helper method for cloning the map. Returns either the default value or
--   the reference to the factory function.
____exports.DefaultMap = __TS__Class()
local DefaultMap = ____exports.DefaultMap
DefaultMap.name = "DefaultMap"
__TS__ClassExtends(DefaultMap, Map)
function DefaultMap.prototype.____constructor(self, defaultValueOrFactoryFunction, initializerArray)
    local argIsPrimitive = isPrimitive(nil, defaultValueOrFactoryFunction)
    local argIsFunction = isFunction(nil, defaultValueOrFactoryFunction)
    if not argIsPrimitive and not argIsFunction then
        error(("Failed to instantiate a DefaultMap since the provided default value was of type \"" .. __TS__TypeOf(defaultValueOrFactoryFunction)) .. "\". This error usually means that you are trying to use an array (or some other non-primitive data structure that is passed by reference) as the default value. Instead, return the data structure in a factory function, like \"() => []\". See the DefaultMap documentation for more details.")
    end
    Map.prototype.____constructor(self, initializerArray)
    if argIsFunction then
        self.defaultValue = nil
        self.defaultValueFactory = defaultValueOrFactoryFunction
    else
        self.defaultValue = defaultValueOrFactoryFunction
        self.defaultValueFactory = nil
    end
end
function DefaultMap.prototype.getAndSetDefault(self, key, ...)
    local value = Map.prototype.get(self, key)
    if value ~= nil then
        return value
    end
    local defaultValue = self:getDefaultValue(...)
    self:set(key, defaultValue)
    return defaultValue
end
function DefaultMap.prototype.getDefaultValue(self, ...)
    if self.defaultValue ~= nil then
        return self.defaultValue
    end
    if self.defaultValueFactory ~= nil then
        return self:defaultValueFactory(...)
    end
    error("A DefaultMap was incorrectly instantiated.")
end
function DefaultMap.prototype.getConstructorArg(self)
    if self.defaultValue ~= nil then
        return self.defaultValue
    end
    if self.defaultValueFactory ~= nil then
        return self.defaultValueFactory
    end
    error("A DefaultMap was incorrectly instantiated.")
end
local function test(self)
    local myDefaultMapBoolean = __TS__New(____exports.DefaultMap, false)
    local myDefaultMapBooleanFactory = __TS__New(
        ____exports.DefaultMap,
        function() return false end
    )
    local myDefaultMapBooleanWithoutParams = __TS__New(____exports.DefaultMap, false)
    local myDefaultMapNumber = __TS__New(____exports.DefaultMap, 123)
    local myDefaultMapNumberFactory = __TS__New(
        ____exports.DefaultMap,
        function() return 123 end
    )
    local myDefaultMapNumberWithoutParams = __TS__New(____exports.DefaultMap, 123)
    local myDefaultMapString = __TS__New(____exports.DefaultMap, "foo")
    local myDefaultMapStringFactory = __TS__New(
        ____exports.DefaultMap,
        function() return "foo" end
    )
    local myDefaultMapStringWithoutParams = __TS__New(____exports.DefaultMap, "foo")
    local myDefaultMapArray = __TS__New(
        ____exports.DefaultMap,
        function() return {} end
    )
    local myDefaultMapArrayWithoutParams = __TS__New(
        ____exports.DefaultMap,
        function() return {} end
    )
    local myDefaultMapMap = __TS__New(
        ____exports.DefaultMap,
        function() return __TS__New(Map) end
    )
    local myDefaultMapMapWithoutParams = __TS__New(
        ____exports.DefaultMap,
        function() return __TS__New(Map) end
    )
end
return ____exports
